@@ -20,6 +20,12 @@
 #include <stdlib.h>
 #include <sslproto.h>
 
+/* Cipher actions */
+#define PERMANENTLY_DISABLE_CIPHER   -1 /* !CIPHER */
+#define SUBTRACT_CIPHER               0 /* -CIPHER */
+#define ENABLE_CIPHER                 1 /* CIPHER */
+#define REORDER_CIPHER                2 /* +CIPHER */
+
 /* ciphernum is defined in nss_engine_cipher.h */
 cipher_properties ciphers_def[] =
 {
@@ -32,6 +38,9 @@ cipher_properties ciphers_def[] =
     /* TLS_RSA_EXPORT_WITH_DES40_CBC_SHA not implemented 0x0008 */
     {"rsa_des_sha", TLS_RSA_WITH_DES_CBC_SHA, "DES-CBC-SHA", SSL_kRSA|SSL_aRSA|SSL_DES|SSL_SHA1, SSLV3, SSL_LOW, 56, 56},
     {"rsa_3des_sha", TLS_RSA_WITH_3DES_EDE_CBC_SHA, "DES-CBC3-SHA", SSL_kRSA|SSL_aRSA|SSL_3DES|SSL_SHA1, SSLV3, SSL_HIGH, 168, 168},
+#ifdef ENABLE_SERVER_DHE
+    {"dhe_rsa_des_sha", TLS_DHE_RSA_WITH_DES_CBC_SHA, "EDH-RSA-DES-CBC-SHA", SSL_kEDH|SSL_aRSA|SSL_DES|SSL_SHA1, SSLV3, SSL_LOW, 56, 56},
+#endif
     {"rsa_aes_128_sha", TLS_RSA_WITH_AES_128_CBC_SHA, "AES128-SHA", SSL_kRSA|SSL_aRSA|SSL_AES128|SSL_SHA1, TLSV1, SSL_HIGH, 128, 128},
     {"rsa_aes_256_sha", TLS_RSA_WITH_AES_256_CBC_SHA, "AES256-SHA", SSL_kRSA|SSL_aRSA|SSL_AES256|SSL_SHA1, TLSV1, SSL_HIGH, 256, 256},
     {"null_sha_256", TLS_RSA_WITH_NULL_SHA256, "NULL-SHA256", SSL_kRSA|SSL_aRSA|SSL_eNULL|SSL_SHA256, TLSV1_2, SSL_STRONG_NONE, 0, 0},
@@ -49,6 +58,21 @@ cipher_properties ciphers_def[] =
 #endif
     {"fips_3des_sha", SSL_RSA_FIPS_WITH_3DES_EDE_CBC_SHA, "FIPS-DES-CBC3-SHA", SSL_kRSA|SSL_aRSA|SSL_3DES|SSL_SHA1, SSLV3, SSL_HIGH, 112, 168},
     {"fips_des_sha", SSL_RSA_FIPS_WITH_DES_CBC_SHA, "FIPS-DES-CBC-SHA", SSL_kRSA|SSL_aRSA|SSL_DES|SSL_SHA1, SSLV3, SSL_LOW, 56, 56},
+#ifdef ENABLE_SERVER_DHE
+    {"dhe_rsa_3des_sha", TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA, "EDH-RSA-DES-CBC3-SHA", SSL_kEDH|SSL_aRSA|SSL_3DES|SSL_SHA1, TLSV1, SSL_HIGH, 112, 168},
+    {"dhe_rsa_aes_128_sha", TLS_DHE_RSA_WITH_AES_128_CBC_SHA, "DHE-RSA-AES128-SHA", SSL_kEDH|SSL_aRSA|SSL_AES128|SSL_SHA1, TLSV1, SSL_HIGH, 128, 128},
+    {"dhe_rsa_aes_256_sha", TLS_DHE_RSA_WITH_AES_256_CBC_SHA, "DHE-RSA-AES256-SHA", SSL_kEDH|SSL_aRSA|SSL_AES256|SSL_SHA1, TLSV1, SSL_HIGH, 256, 256},
+    {"dhe_rsa_camellia_128_sha", TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA, "DHE-RSA-CAMELLIA128-SHA", SSL_kEDH|SSL_aRSA|SSL_CAMELLIA128|SSL_SHA1, TLSV1, SSL_HIGH, 128, 128},
+    {"dhe_rsa_camellia_256_sha", TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA, "DHE-RSA-CAMELLIA256-SHA", SSL_kEDH|SSL_aRSA|SSL_CAMELLIA256|SSL_SHA1, TLSV1, SSL_HIGH, 256, 256},
+    {"dhe_rsa_aes_128_sha256", TLS_DHE_RSA_WITH_AES_128_CBC_SHA256, "DHE-RSA-AES128-SHA256", SSL_kEDH|SSL_aRSA|SSL_AES128|SSL_SHA256, TLSV1_2, SSL_HIGH, 128, 128},
+    {"dhe_rsa_aes_256_sha256", TLS_DHE_RSA_WITH_AES_256_CBC_SHA256, "DHE-RSA-AES256-SHA256", SSL_kEDH|SSL_aRSA|SSL_AES256|SSL_SHA256, TLSV1_2, SSL_HIGH, 256, 256},
+#ifdef ENABLE_GCM
+    {"dhe_rsa_aes_128_gcm_sha_256", TLS_DHE_RSA_WITH_AES_128_GCM_SHA256, "DHE-RSA-AES128-GCM-SHA256", SSL_kEDH|SSL_aRSA|SSL_AES128GCM|SSL_AEAD, TLSV1_2, SSL_HIGH, 128, 128},
+#endif
+#ifdef ENABLE_SHA384
+    {"dhe_rsa_aes_256_gcm_sha_384", TLS_DHE_RSA_WITH_AES_256_GCM_SHA384, "DHE-RSA-AES256-GCM-SHA384", SSL_kEDH|SSL_aRSA|SSL_AES256GCM|SSL_AEAD, TLSV1_2, SSL_HIGH, 256, 256},
+#endif
+#endif /* ENABLE_SERVER_DHE */
 #ifdef NSS_ENABLE_ECC
     {"ecdh_ecdsa_null_sha", TLS_ECDH_ECDSA_WITH_NULL_SHA, "ECDH-ECDSA-NULL-SHA", SSL_kECDHe|SSL_aECDH|SSL_eNULL|SSL_SHA1, TLSV1, SSL_STRONG_NONE, 0, 0},
     {"ecdh_ecdsa_rc4_128_sha", TLS_ECDH_ECDSA_WITH_RC4_128_SHA, "ECDH-ECDSA-RC4-SHA", SSL_kECDHe|SSL_aECDH|SSL_RC4|SSL_SHA1, TLSV1, SSL_MEDIUM, 128, 128},
@@ -115,7 +139,7 @@ static int parse_openssl_ciphers(server_rec *s, char *ciphers, PRBool cipher_lis
 
 int countciphers(PRBool cipher_state[ciphernum], int version) {
     int ciphercount = 0;
-    int i = ciphernum;
+    int i = 0;
 
     for (i = 0; i < ciphernum; i++)
     {
@@ -163,14 +187,18 @@ static void set_cipher_value(PRBool cipher_list[ciphernum], int index, int actio
 {
     int i;
 
+    if (action == REORDER_CIPHER)
+        /* NSS doesn't allow ordering so do nothing */
+        return;
+
     for (i = 0; i < skip_ciphers; i++) {
         if (ciphers_def[index].num == ciphers_not_in_openssl[i]) {
-            cipher_list[index] = -1;
+            cipher_list[index] = PERMANENTLY_DISABLE_CIPHER;
             return;
         }
     }
 
-    if (cipher_list[index] != -1) /* cipher is disabled */
+    if (cipher_list[index] != PERMANENTLY_DISABLE_CIPHER)
         cipher_list[index] = action;
 }
 
@@ -189,23 +217,24 @@ static int parse_openssl_ciphers(server_rec *s, char *ciphers, PRBool cipher_lis
         while ((*cipher) && (isspace(*cipher)))
             ++cipher;
 
-        action = 1; /* default to enable */
+        action = ENABLE_CIPHER; /* default to enable */
         switch(*cipher)
         {
-            case '+': /* Add something */
+            case '+':
                 /* Cipher ordering is not supported in NSS */
-                return 0;
-                break;
-            case '-': /* Subtract something */
-                action = 0;
+                action = REORDER_CIPHER;
                 cipher++;
                 break;
-            case '!':  /* Disable something */
-                action = -1;
+            case '-':
+                action = SUBTRACT_CIPHER;
+                cipher++;
+                break;
+            case '!':
+                action = PERMANENTLY_DISABLE_CIPHER;
                 cipher++;
                 break;
             default:
-               /* do nothing */
+                /* Add the cipher */
                 break;
         }
 
@@ -235,12 +264,13 @@ static int parse_openssl_ciphers(server_rec *s, char *ciphers, PRBool cipher_lis
             int mask = SSL_aNULL | SSL_eNULL;
             found = PR_TRUE;
             for (i=0; i < ciphernum; i++) {
-                if (cipher_list[i] != -1)
+                if (cipher_list[i] != PERMANENTLY_DISABLE_CIPHER)
                     SSL_CipherPrefGetDefault(ciphers_def[i].num,
                                              &cipher_list[i]);
                 if (PR_TRUE == first) {
                     if (ciphers_def[i].attr & mask) {
-                        set_cipher_value(cipher_list, i, -1);
+                        set_cipher_value(cipher_list, i,
+                                         PERMANENTLY_DISABLE_CIPHER);
                     }
                 }
             }
@@ -289,7 +319,16 @@ static int parse_openssl_ciphers(server_rec *s, char *ciphers, PRBool cipher_lis
                 } else if (!strcmp(cipher, "aRSA")) {
                     mask |= SSL_aRSA;
                 } else if (!strcmp(cipher, "EDH")) {
-                    mask |= SSL_EDH;
+                    /* Normally this is kEDH:-ADH but since we don't
+                     * support ADH this is sufficient.
+                     */
+                    mask |= SSL_kEDH;
+                } else if (!strcmp(cipher, "DH")) {
+                    /* non-ephemeral DH. The ciphers are defined
+                     * but not implemented in OpenSSL so manage
+                     * this here.
+                     */
+                    mask |= SSL_kEDH;
 #if 0
                 } else if (!strcmp(cipher, "ADH")) {
                     mask |= SSL_ADH;
@@ -387,7 +426,7 @@ static int parse_openssl_ciphers(server_rec *s, char *ciphers, PRBool cipher_lis
                         if (((ciphers_def[i].attr & mask) ||
                          (ciphers_def[i].strength & strength) ||
                          (ciphers_def[i].version & protocol)) &&
-                         (cipher_list[i] != -1)) {
+                         (cipher_list[i] != PERMANENTLY_DISABLE_CIPHER)) {
                             if (amask != 0) {
                                 PRBool match = PR_FALSE;
                                 if (ciphers_def[i].attr & amask) {
