@@ -27,6 +27,8 @@
 #include "ocsp.h"
 #include "keyhi.h"
 #include "cert.h"
+#include <sys/types.h>
+#include <pwd.h>
 
 static SECStatus ownBadCertHandler(void *arg, PRFileDesc * socket);
 static SECStatus ownHandshakeCallback(PRFileDesc * socket, void *arg);
@@ -49,6 +51,32 @@ static char *version_components[] = {
     NULL
 };
 
+/* See if a uid or gid can read a file at a given path. Ignore world
+ * read permissions.
+ *
+ * Return 0 on failure or file doesn't exist
+ * Return 1 on success
+ */
+static int check_path(uid_t uid, gid_t gid, char *filepath, apr_pool_t *p)
+{
+    apr_finfo_t finfo;
+    int rv;
+
+    if ((rv = apr_stat(&finfo, filepath, APR_FINFO_PROT | APR_FINFO_OWNER,
+         p)) == APR_SUCCESS) {
+        if (((uid == finfo.user) &&
+            ((finfo.protection & APR_FPROT_UREAD))) ||
+            ((gid == finfo.group) &&
+                ((finfo.protection & APR_FPROT_GREAD)))
+           )
+        {
+            return 1;
+        }
+        return 0;
+    }
+    return 0;
+}
+
 static char *nss_add_version_component(apr_pool_t *p,
                                        server_rec *s,
                                        char *name)
@@ -130,6 +158,68 @@ static void nss_init_SSLLibrary(server_rec *base_server, apr_pool_t *p)
         }
     }
 
+    /* Assuming everything is ok so far, check the cert database permissions
+     * for the server user before Apache starts forking. We die now or
+     * get stuck in an endless loop not able to read the NSS database.
+     */
+    if (mc->nInitCount == 1) {
+        if (mc->skip_permission_check == PR_FALSE) {
+            char filepath[1024];
+            struct passwd *pw = NULL;
+
+            pw = getpwnam(mc->user);
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, base_server,
+                "Checking permissions for user %s: uid %d gid %d",
+                mc->user, pw->pw_uid, pw->pw_gid);
+
+            if (strncasecmp(mc->pCertificateDatabase, "sql:", 4) == 0) {
+                apr_snprintf(filepath, 1024, "%s/key4.db",
+                             mc->pCertificateDatabase+4);
+                if (!(check_path(pw->pw_uid, pw->pw_gid, filepath, p))) {
+                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+                        "Server user %s lacks read access to NSS key "
+                        "database %s.", mc->user, filepath);
+                    nss_die();
+                }
+                apr_snprintf(filepath, 1024, "%s/cert9.db",
+                             mc->pCertificateDatabase+4);
+                if (!(check_path(pw->pw_uid, pw->pw_gid, filepath, p))) {
+                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+                        "Server user %s lacks read access to NSS cert "
+                        "database %s.", mc->user, filepath);
+                    nss_die();
+                }
+            } else {
+                apr_snprintf(filepath, 1024, "%s/key3.db",
+                             mc->pCertificateDatabase);
+                if (!(check_path(pw->pw_uid, pw->pw_gid, filepath, p))) {
+                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+                        "Server user %s lacks read access to NSS key "
+                        "database %s.", mc->user, filepath);
+                    nss_die();
+                }
+                apr_snprintf(filepath, 1024, "%s/cert8.db",
+                             mc->pCertificateDatabase);
+                if (!(check_path(pw->pw_uid, pw->pw_gid, filepath, p))) {
+                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+                        "Server user %s lacks read access to NSS cert "
+                        "database %s.", mc->user, filepath);
+                    nss_die();
+                }
+                apr_snprintf(filepath, 1024, "%s/secmod.db",
+                             mc->pCertificateDatabase);
+                if (!(check_path(pw->pw_uid, pw->pw_gid, filepath, p))) {
+                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+                        "Server user %s lacks read access to NSS secmod "
+                        "database %s.", mc->user, filepath);
+                    nss_die();
+                }
+            }
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, base_server,
+                "Permissions check out ok");
+        }
+    }
+
     /* We need to be in the same directory as libnssckbi.so to load the
      * root certificates properly.
      */
@@ -155,6 +245,7 @@ static void nss_init_SSLLibrary(server_rec *base_server, apr_pool_t *p)
         else
             return;
     }
+
     /* Initialize NSS and open the certificate database read-only. */
     rv = NSS_Initialize(mc->pCertificateDatabase, mc->pDBPrefix, mc->pDBPrefix, "secmod.db", NSS_INIT_READONLY);
     if (chdir(cwd) != 0) {
@@ -168,39 +259,16 @@ static void nss_init_SSLLibrary(server_rec *base_server, apr_pool_t *p)
             return;
     }
 
-    /* Assuming everything is ok so far, check the cert database password(s). */
     if (rv != SECSuccess) {
-        apr_finfo_t finfo;
-        char keypath[1024];
-        int rv;
-        uid_t user_id;
-        gid_t gid;
-
-        user_id = ap_uname2id(mc->user);
-        gid = getegid();
-
         NSS_Shutdown();
+
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
             "NSS_Initialize failed. Certificate database: %s.", mc->pCertificateDatabase != NULL ? mc->pCertificateDatabase : "not set in configuration");
 
         nss_log_nss_error(APLOG_MARK, APLOG_ERR, base_server);
-        apr_snprintf(keypath, 1024, "%s/key3.db", mc->pCertificateDatabase);
-        if ((rv = apr_stat(&finfo, keypath, APR_FINFO_PROT | APR_FINFO_OWNER,
-             p)) == APR_SUCCESS) {
-            if (((user_id == finfo.user) &&
-                    (!(finfo.protection & APR_FPROT_UREAD))) ||
-                ((gid == finfo.group) &&
-                    (!(finfo.protection & APR_FPROT_GREAD))) ||
-                (!(finfo.protection & APR_FPROT_WREAD))
-               )
-            {
-                ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
-                    "Server user lacks read access to NSS database.");
-            }
-        } else {
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
-                "Does the NSS database exist?");
-        }
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+            "Does the NSS database exist?");
+
         nss_die();
     }
 
@@ -371,9 +439,8 @@ int nss_init_Module(apr_pool_t *p, apr_pool_t *plog,
          * Create the server host:port string because we need it a lot
          */
         sc->vhost_id = nss_util_vhostid(p, s);
-        sc->vhost_id_len = strlen(sc->vhost_id);
 
-        if (sc->sni && sc->server->nickname != NULL && sc->vhost_id != NULL) {
+        if (sc->sni && sc->server && sc->server->nickname != NULL && sc->vhost_id != NULL) {
             split_vhost_id = apr_strtok((char *)sc->vhost_id, ":", &last1);
             ap_str_tolower(split_vhost_id);
             addHashVhostNick(split_vhost_id, (char *)sc->server->nickname);
@@ -649,77 +716,86 @@ static void nss_init_ctx_protocol(server_rec *s,
         protocol_marker = "NSSProxyProtocol";
     }
 
-    if (mctx->sc->fips) {
-        ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
-            "In FIPS mode ignoring %s list, enabling TLSv1.0, TLSv1.1 and TLSv1.2",
+    if (mctx->auth.protocols == NULL) {
+        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+            "%s value not set; using: TLSv1.0, TLSv1.1 and TLSv1.2",
             protocol_marker);
         tls = tls1_1 = tls1_2 = 1;
     } else {
-        if (mctx->auth.protocols == NULL) {
-            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
-                "%s value not set; using: TLSv1.0, TLSv1.1 and TLSv1.2",
-                protocol_marker);
-            tls = tls1_1 = tls1_2 = 1;
+        lprotocols = strdup(mctx->auth.protocols);
+        ap_str_tolower(lprotocols);
+
+        if (strstr(lprotocols, "all") != NULL) {
+            ssl3 = tls = tls1_1 = tls1_2 = 1;
         } else {
-            lprotocols = strdup(mctx->auth.protocols);
-            ap_str_tolower(lprotocols);
+            char *protocol_list = NULL;
+            char *saveptr = NULL;
+            char *token = NULL;
 
-            if (strstr(lprotocols, "all") != NULL) {
-                ssl3 = tls = tls1_1 = tls1_2 = 1;
-            } else {
-                char *protocol_list = NULL;
-                char *saveptr = NULL;
-                char *token = NULL;
-
-                for (protocol_list = lprotocols; ; protocol_list = NULL) {
-                    token = strtok_r(protocol_list, ",", &saveptr);
-                    if (token == NULL) {
-                        break;
-                    } else if (strcmp(token, "sslv2") == 0) {
-                        ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
-                                     "%s:  SSL2 is not supported",
+            for (protocol_list = lprotocols; ; protocol_list = NULL) {
+                token = strtok_r(protocol_list, ",", &saveptr);
+                if (token == NULL) {
+                    break;
+                } else if (strcmp(token, "sslv2") == 0) {
+                    ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
+                                 "%s:  SSL2 is not supported",
+                                 protocol_marker);
+                } else if (strcmp(token, "sslv3") == 0) {
+                    if (mctx->sc->fips) {
+                        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+                                     "%s: SSL3 is disabled by FIPS policy",
                                      protocol_marker);
-                    } else if (strcmp(token, "sslv3") == 0) {
+                    } else {
                         ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                                      "%s:  Enabling SSL3",
                                      protocol_marker);
                         ssl3 = 1;
-                    } else if (strcmp(token, "tlsv1") == 0) {
-                        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
-                                     "%s:  Enabling TLSv1.0 via TLSv1",
-                                     protocol_marker);
-                        ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
-                                     "%s:  The 'TLSv1' protocol name has been deprecated; please change 'TLSv1' to 'TLSv1.0'.",
-                                     protocol_marker);
-                        tls = 1;
-                    } else if (strcmp(token, "tlsv1.0") == 0) {
-                        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
-                                     "%s:  Enabling TLSv1.0",
-                                     protocol_marker);
-                        tls = 1;
-                    } else if (strcmp(token, "tlsv1.1") == 0) {
-                        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
-                                     "%s:  Enabling TLSv1.1",
-                                     protocol_marker);
-                        tls1_1 = 1;
-                    } else if (strcmp(token, "tlsv1.2") == 0) {
-                        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
-                                     "%s:  Enabling TLSv1.2",
-                                     protocol_marker);
-                        tls1_2 = 1;
-                    } else {
-                        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
-                                     "%s:  Unknown protocol '%s' not supported",
-                                     protocol_marker, token);
                     }
+                } else if (strcmp(token, "tlsv1") == 0) {
+                    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                                 "%s:  Enabling TLSv1.0 via TLSv1",
+                                 protocol_marker);
+                    ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
+                                 "%s:  The 'TLSv1' protocol name has been deprecated; please change 'TLSv1' to 'TLSv1.0'.",
+                                 protocol_marker);
+                    tls = 1;
+                } else if (strcmp(token, "tlsv1.0") == 0) {
+                    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                                 "%s:  Enabling TLSv1.0",
+                                 protocol_marker);
+                    tls = 1;
+                } else if (strcmp(token, "tlsv1.1") == 0) {
+                    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                                 "%s:  Enabling TLSv1.1",
+                                 protocol_marker);
+                    tls1_1 = 1;
+                } else if (strcmp(token, "tlsv1.2") == 0) {
+                    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                                 "%s:  Enabling TLSv1.2",
+                                 protocol_marker);
+                    tls1_2 = 1;
+                } else {
+                    ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+                                 "%s:  Unknown protocol '%s' not supported",
+                                 protocol_marker, token);
                 }
             }
-            free(lprotocols);
+        }
+        free(lprotocols);
+
+        /*
+         * After processing the specified protocols list,
+         * if FIPS mode is enabled with no TLS protocols,
+         * enable ALL TLS protocols.
+         */
+        if ((mctx->sc->fips) && (tls == 0) && (tls1_1 == 0) && (tls1_2 == 0)) {
+            ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
+                "%s: FIPS mode no valid protocols set, enabling TLSv1.0, TLSv1.1 and TLSv1.2",
+                protocol_marker);
+            tls = tls1_1 = tls1_2 = 1;
         }
     }
 
-    stat = SECSuccess;
-
     stat = SSL_OptionSet(mctx->model, SSL_ENABLE_SSL2, PR_FALSE);
 
     /* Set protocol version ranges:
@@ -829,6 +905,17 @@ static void nss_init_ctx_protocol(server_rec *s,
         nss_log_nss_error(APLOG_MARK, APLOG_ERR, s);
         nss_die();
     }
+#ifdef ENABLE_SERVER_DHE
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+        "Enabling DHE key exchange");
+    if (SSL_OptionSet(mctx->model, SSL_ENABLE_SERVER_DHE,
+        PR_TRUE) != SECSuccess) {
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
+                "Unable to enable DHE key exchange");
+        nss_log_nss_error(APLOG_MARK, APLOG_ERR, s);
+        nss_die();
+    }
+#endif
 }
 
 static void nss_init_ctx_session_cache(server_rec *s,
@@ -1043,6 +1130,10 @@ static void nss_init_ctx_cipher_suite(server_rec *s,
 
     /* Finally actually enable the selected ciphers */
     for (i=0; i<ciphernum;i++) {
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+            "%sable cipher: %s",
+            cipher_state[i] == 1 ? "En" : "Dis",
+            ciphers_def[i].name);
         SSL_CipherPrefSet(mctx->model, ciphers_def[i].num, cipher_state[i] == 1 ? PR_TRUE : PR_FALSE);
     }
 }
@@ -1236,6 +1327,7 @@ static void nss_init_certificate(server_rec *s, const char *nickname,
         case secCertTimeNotValidYet:
             ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                 "Certificate is not valid yet '%s'", nickname);
+            break;
         default:
             ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                 "Unhandled Certificate time type %d for: '%s'", certtimestatus, nickname);
@@ -1281,9 +1373,8 @@ static void nss_init_server_certs(server_rec *s,
 #endif
         {
             /*
-             * Since this is a 'fatal' error, regardless of whether this
-             * particular invocation is from a 'server' object or a 'proxy'
-             * object, issue all error message(s) as appropriate.
+             * This is a 'fatal' error since a 'nickname' is required for
+             * a 'server' object; log error message(s) as appropriate.
              */
             if ((mctx->sc->enabled == TRUE) &&
                 (mctx->sc->server) &&
@@ -1292,13 +1383,6 @@ static void nss_init_server_certs(server_rec *s,
                     "NSSEngine on; no certificate nickname provided by NSSNickname.");
             }
 
-            if ((mctx->sc->proxy_enabled == TRUE) &&
-                (mctx->sc->proxy) &&
-                (mctx->sc->proxy->nickname == NULL)) {
-                ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-                    "NSSProxyEngine on; no certificate nickname provided by NSSProxyNickname.");
-            }
-
             nss_die();
         }
 
@@ -1327,6 +1411,11 @@ static void nss_init_server_certs(server_rec *s,
             "SSL error configuring handshake callback: '%s'", mctx->nickname);
         nss_log_nss_error(APLOG_MARK, APLOG_ERR, s);
         nss_die();
+    } else {
+        if (!mctx->as_server && mctx->nickname != NULL) {
+            ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
+                "Configured proxy nickname as '%s'", mctx->nickname);
+        }
     }
 
 }
@@ -1448,7 +1537,7 @@ apr_status_t nss_init_ModuleKill(void *data)
     server_rec *base_server = (server_rec *)data;
     SSLModConfigRec *mc = myModConfig(base_server);
 
-    if (!NSS_IsInitialized()) {
+    if (!NSS_IsInitialized() && !PR_Initialized()) {
         return APR_SUCCESS;
     }
 
@@ -1483,7 +1572,7 @@ apr_status_t nss_init_ChildKill(void *data)
     for (s = base_server; s; s = s->next) {
         sc = mySrvConfig(s);
 
-        if (sc->enabled == TRUE) {
+        if (sc->enabled == TRUE && NSS_IsInitialized()) {
             if (sc->server->nickname) {
                 CERT_DestroyCertificate(sc->server->servercert);
                 SECKEY_DestroyPrivateKey(sc->server->serverkey);
@@ -1502,7 +1591,7 @@ apr_status_t nss_init_ChildKill(void *data)
 
             shutdown = 1;
         }
-        if (sc->proxy_enabled) {
+        if (sc->proxy_enabled && NSS_IsInitialized()) {
             if (sc->proxy->servercert != NULL) {
                 CERT_DestroyCertificate(sc->proxy->servercert);
                 SECKEY_DestroyPrivateKey(sc->proxy->serverkey);
@@ -1731,9 +1820,6 @@ PRInt32 nssSSLSNISocketConfig(PRFileDesc *fd, const SECItem *sniNameArr,
 {
     server_rec *s = (server_rec *)arg;
 
-    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
-		 "nssSSLSNISocketConfig");
-
     void *pinArg;
     CERTCertificate *cert = NULL;
     SECKEYPrivateKey *privKey = NULL;
@@ -1758,8 +1844,25 @@ PRInt32 nssSSLSNISocketConfig(PRFileDesc *fd, const SECItem *sniNameArr,
         nickName = searchHashVhostbyNick_match(vhost);
         if (nickName == NULL) {
             ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
-                "SNI: Search for %s failed. Unrecognized name.", vhost);
-            goto loser;
+                "SNI: No matching SSL virtual host for servername "
+                "%s found (using default/first virtual host)",
+                vhost);
+            /*
+             * RFC 6066 section 3 says "It is NOT RECOMMENDED to send
+             * a warning-level unrecognized_name(112) alert, because
+             * the client's behavior in response to warning-level alerts
+             * is unpredictable."
+             *
+             * To maintain compatibility with mod_ssl, we won't send
+             * any alert (neither warning- nor fatal-level),
+             * i.e. we take the second action suggested in RFC 6066:
+             * "If the server understood the ClientHello extension but
+             * does not recognize the server name, the server SHOULD take
+             * one of two actions: either abort the handshake by sending
+             * a fatal-level unrecognized_name(112) alert or continue
+             * the handshake."
+             */
+             return 0;
         }
     }
 
