@@ -150,9 +150,9 @@ int nss_hook_ReadReq(request_rec *r)
      * Log information about incoming HTTPS requests
      */
 #if AP_SERVER_MINORVERSION_NUMBER <= 2
-    if (r->server->loglevel >= APLOG_INFO && ap_is_initial_req(r)) {
+    if (r->server && r->server->loglevel >= APLOG_INFO && ap_is_initial_req(r)) {
 #else
-    if (r->server->log.level >= APLOG_INFO && ap_is_initial_req(r)) {
+    if (r->server && r->server->log.level >= APLOG_INFO && ap_is_initial_req(r)) {
 #endif
         ap_log_error(APLOG_MARK, APLOG_INFO, 0, r->server,
                      "%s HTTPS request received for child %ld (server %s)",
@@ -198,7 +198,7 @@ int nss_hook_Access(request_rec *r)
     /*
      * Support for SSLRequireSSL directive
      */
-    if (dc->bSSLRequired && !ssl) {
+    if (dc->bSSLRequired && (!ssl || !sslconn)) {
         ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                       "access to %s failed, reason: %s",
                       r->filename, "SSL connection required");
@@ -218,6 +218,11 @@ int nss_hook_Access(request_rec *r)
         return DECLINED;
     }
 
+    /* prevent NULL dereferences */
+    if (!sslconn) {
+        return DECLINED;
+    }
+
     /*
      * Support for per-directory reconfigured SSL connection parameters.
      *
@@ -273,6 +278,8 @@ int nss_hook_Access(request_rec *r)
             SSL_SecurityStatus(ssl, &on, &cipher,
                                &keySize, &secretKeySize, &issuer,
                                &subject);
+            PORT_Free(issuer);
+            PORT_Free(subject);
         }
 
         /* configure new state */
@@ -577,12 +584,18 @@ int nss_hook_Access(request_rec *r)
             int on, keySize, secretKeySize;
             char *issuer, *subject;
 
+            PORT_Free(cipher);
+
             SSL_SecurityStatus(ssl, &on, &cipher,
                                &keySize, &secretKeySize, &issuer,
                                &subject);
 
             ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
                 "Re-negotiated cipher %s", cipher);
+
+            PORT_Free(cipher);
+            PORT_Free(issuer);
+            PORT_Free(subject);
         }
 
         /*
@@ -607,8 +620,25 @@ int nss_hook_Access(request_rec *r)
     if ((dc->nOptions & SSL_OPT_FAKEBASICAUTH) == 0 && dc->szUserName) {
         char *val = nss_var_lookup(r->pool, r->server, r->connection,
                                    r, (char *)dc->szUserName);
-        if (val && val[0])
-            r->user = val;
+        if (val && val[0]) {
+            /* RFC2617 denies usage of colon in BasicAuth */
+            char *colon = strchr(val, ':');
+            if (colon == NULL) {
+                r->user = val;
+            }
+            else {
+                cp = apr_psprintf(r->pool,
+                                "FakeBasicAuth is configured and colon "
+                                "(\":\") character exists in the \"%s\" "
+                                "username", val);
+
+                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                            "access to %s failed, reason: %s",
+                            r->filename, cp);
+
+                return HTTP_FORBIDDEN;
+            }
+        }
     }
 
     /*
@@ -753,6 +783,20 @@ int nss_hook_UserCheck(request_rec *r)
 
     clientdn = (char *)sslconn->client_dn;
 
+    char *colon = strchr(clientdn, ':');
+    if (colon != NULL) {
+        char *cp = apr_psprintf(r->pool,
+                        "FakeBasicAuth is configured and colon "
+                        "(\":\") character exists in the \"%s\" "
+                        "username", clientdn);
+
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                    "access to %s failed, reason: %s",
+                    r->filename, cp);
+
+        return HTTP_FORBIDDEN;
+    }
+
     /*
      * Fake a password - which one would be immaterial, as, it seems, an empty
      * password in the users file would match ALL incoming passwords, if only
@@ -933,6 +977,7 @@ int nss_hook_Fixup(request_rec *r)
         SECITEM_FreeItem(hostInfo, PR_TRUE);
     }
 
+    modnss_var_extract_san_entries(env, sslconn->ssl, r->pool);
     /* standard SSL environment variables */
     if (dc->nOptions & SSL_OPT_STDENVVARS) {
         for (i = 0; nss_hook_Fixup_vars[i]; i++) {
