@@ -211,7 +211,11 @@ CreatePk11PinStore(Pk11PinStore **out, const char *tokenName, const char *pin)
 
             ctx = PK11_CreateContextBySymKey(store->mech->type, CKA_ENCRYPT,
                     store->key, store->params);
-            if (!ctx) { err = PIN_SYSTEMERROR; break; }
+            if (!ctx) {
+                err = PIN_SYSTEMERROR;
+                free(plain);
+                break;
+            }
 
             do {
                 rv = PK11_CipherOp(ctx, store->crypt, &outLen, store->length,
@@ -383,8 +387,6 @@ int main(int argc, char ** argv)
                 break;
             }
             command = getstr(buf, 0);
-            tokenName = getstr(buf, 1);
-            tokenpw = getstr(buf, 2);
 
             if (command && !strcmp(command, "QUIT")) {
                 break;
@@ -392,17 +394,23 @@ int main(int argc, char ** argv)
                 PRInt32 err = PIN_SUCCESS;
                 Node *node = NULL;
 
+                tokenName = getstr(buf, 1);
+                tokenpw = getstr(buf, 2);
+
                 if (tokenName && tokenpw) {
                     node = (Node*)malloc(sizeof (Node));
-                    if (!node) { err = PIN_NOMEMORY; }
+                    if (!node) {
+                        err = PIN_NOMEMORY;
+                    } else {
+                        node->tokenName = strdup(tokenName);
+                        node->store = 0;
+                        node->next = 0;
 
-                    node->tokenName = strdup(tokenName);
-                    node->store = 0;
-                    node->next = 0;
-
-                    if (err == PIN_SUCCESS)
-                        err = CreatePk11PinStore(&node->store, tokenName, tokenpw);
-                    memset(tokenpw, 0, strlen(tokenpw));
+                        if (err == PIN_SUCCESS)
+                            err = CreatePk11PinStore(&node->store,
+                                                     tokenName, tokenpw);
+                        memset(tokenpw, 0, strlen(tokenpw));
+                    }
                 } else
                     err = PIN_SYSTEMERROR;
 
@@ -424,6 +432,8 @@ int main(int argc, char ** argv)
                 char *pin = 0;
                 PRBool found = PR_FALSE;
 
+                tokenName = getstr(buf, 1);
+
                 for (node = pinList; node != NULL; node = node->next) {
                     if (!strcmp(node->tokenName, tokenName)) {
                         if (Pk11StoreGetPin(&pin, node->store) == SECSuccess) {
@@ -468,7 +478,12 @@ char * getstr(const char * cmd, int el) {
     char *peek;
     int i = 0;
 
+    if (cmd == NULL)
+        return NULL;
+
     work = strdup(cmd);
+    if (!work)
+        return NULL;
     s = t = work;
     r = NULL;
 
